#include "Mouse.h"



// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse Health Science Center

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================



long Mouse::static_ID_Counter = 0; // <-- this has to be reset whenever we clear the purebreds, or everything breaks

Mouse::Mouse(string name, long generation, long ID, Mouse *motherPtr, Mouse *fatherPtr, char sex) {
    // Anything except for name (which must be 'u' 'f' or 'm') that is 
    // passed in as "Mouse::kUNDEFINED" (which is just -1) will be auto-assigned
    
    // ========== NAME
    myName = name;
    
    // ========== GENERATION
    if (generation == kUNDEFINED) {
	// A mouse's generation should be the farthest distance from it to the farthest purebred mouse.
	if (motherPtr != NULL && fatherPtr != NULL)    myGeneration = (1 + max(motherPtr->generation(), fatherPtr->generation()));
	else                                                                      myGeneration = 0;
    } else {
	myGeneration = generation;
    }
    
    // ========== ID
    if (ID == -1)   myID = (static_ID_Counter)++; // sets myID equal to static_ID_Counter, then increments the counter
    else               myID = ID;
    
    // ========== POINTERS
    myMotherPtr = motherPtr;
    myFatherPtr = fatherPtr;
    
    // ========== SEX
    if (sex == 'u') {
        // If the mouse's sex was not assigned, randomly set it
        if (rand() < RAND_MAX/2)   mySex = kMALE; // <-- random assignment
        else                                       mySex = kFEMALE; // <-- random assignment
        
    } else if (sex == 'f') {                  mySex = kFEMALE;
    } else if (sex == 'm') {                  mySex = kMALE;
    } else {
	cerr << "Error in the constructor for Mouse. Passed in a sex that was not \'u\' (unspecified), \'m\', or \'f\'. We passed in the value \'" << sex << "\' instead." << endl;
	assert("Error in constructor for mouse" == 0);
	exit(1);
    }
    
    // ========== INITIALIZE CHROMOSOMES (here, they will be blank--no inheritance is done yet)
    myMatChr.reserve(Chromosome::numChr() - 1); // <-- no Y chr inherited from the mother    
    myPatChr.reserve(Chromosome::numChr());
    
    for (int i = 0; i < Chromosome::numChr(); i++) {
	if (i < Chromosome::yChrIndex()) { myMatChr.push_back(new Chromosome(i)); } // <-- no Y chr on maternal chromosome
	myPatChr.push_back(new Chromosome(i));
    }
    
}

void Mouse::resetID_Counter() {
    static_ID_Counter = 0; // resets the global ID counter
}

void Mouse::setName(string newName) {
    myName = newName;
}

int Mouse::numChr() {
    if (isMale()) { return (Chromosome::numChr()); }
    else return (Chromosome::numChr() - 1);
    // This assumes that whatever animal you are working with has males with
    // chromosomes 1...n + XY, and females have 1...n + XX.
    // So although the animals have the same number of chromosomes, numChr() is actually
    // the number of types of chromosomes. So for males you will get 1..n + X + Y (numChr()), and
    // for females you will get 1..n + X, which is the same as (numChr()-1).
    // If you are working with some animal that doesn't work this way,
    // then this will need to be changed.
}

int Mouse::geneAtBase(bool side, int chrNum, long locInBases) {    
  return getChr(side, chrNum)->getGRfromActual(locInBases)->value();    
}


Chromosome* Mouse::getChr(bool side, int chrNum) {
    if (Mouse::isFemale(side)) {
	if (chrNum >= long(myMatChr.size())) {
	    cerr << "$$$ Error: myMatChr.size() is " << long(myMatChr.size()) << ", but we tried to get an item from index " << chrNum << "." << endl;
	}
	assert(chrNum >= 0 && chrNum < long(myMatChr.size()));
	return myMatChr[chrNum];
	
    } else {
	if (chrNum >= long(myPatChr.size())) {
	    cerr << "$$$ Error: myPatChr.size() is " << long(myPatChr.size()) << ", but we tried to get an item from index " << chrNum << "." << endl;
	}
	
	assert(chrNum >= 0 && chrNum < long(myPatChr.size()));
	return myPatChr[chrNum];
    }
}

void Mouse::setChr(bool side, int chrNum, Chromosome* cPtr) {
    if (Mouse::isFemale(side)) {
	assert(chrNum >= 0 && chrNum < long(myMatChr.size()));
	myMatChr[chrNum] = cPtr;
		
    } else {
	assert(chrNum >= 0 && chrNum < long(myPatChr.size()));
	myPatChr[chrNum] = cPtr;
    }
}

void Mouse::printSummary() {
    cout << "-------------------------\n";
    cout << "NAME: " << myName << endl;
    cout << "ID#: " << myID << ", sex: " << mySex << endl;
    cout << "Number of chromosomes: " << numChr() << endl;
    cout << "Generation: " << myGeneration << endl;
    cout << "Mother's ID: " << myMotherPtr->ID() << endl;
    cout << "Father's ID: " << myFatherPtr->ID() << endl;
    cout << "-------------------------" << endl;
}


void Mouse::printMouseData() {
    // Prints the data about this mouse out to the screen (as text)
    cout << "Mouse: \"" << name() << "\", ID: " << ID()
	    << ", Generation "  << generation() << ", " 
	    <<  numChr() << " chromosomes." << endl;
    
    for (int chr = 0; chr < numChr(); chr++) {
        cout << "Chromosome " << Chromosome::chrNameFromIndex(chr) << ":\n";
        
	if (chr <= Chromosome::xChrIndex()) {
	    // In other words, print out everything from the maternal chr, but don't
	    // try looking for a Y chromosome, because there won't be one.
	    
	    if (getChr(kFEMALE, chr)->numGeneRanges() > 0) {
		cout << "  Maternal Chr (" << getChr(kFEMALE, chr)->numGeneRanges() 
		     << " range(s), "      << getChr(kFEMALE, chr)->lengthBases() << " bases):\n";
	    }
	    
	    for (unsigned int i = 0; i < getChr(kFEMALE, chr)->numGeneRanges(); i++) {
		cout << "     " << getChr(kFEMALE, chr)->getGRfromIndex(i)->value()
		     << ": <"   << getChr(kFEMALE, chr)->getGRfromIndex(i)->start()
		     << " to "  << getChr(kFEMALE, chr)->getGRfromIndex(i)->end() << ">." << endl;
	    }
        }
        if (getChr(kMALE, chr)->numGeneRanges() > 0) {
            cout << "  Paternal Chr (" <<  getChr(kMALE, chr)->numGeneRanges() 
                 << " range(s), "      << getChr(kMALE, chr)->lengthBases() << " bases):\n";
        }
        
        for (unsigned int i = 0; i < getChr(kMALE, chr)->numGeneRanges(); i++) {
            cout << "     " << getChr(kMALE, chr)->getGRfromIndex(i)->value() 
		 << ": <"   << getChr(kMALE, chr)->getGRfromIndex(i)->start() 
		 << " to "  << getChr(kMALE, chr)->getGRfromIndex(i)->end() << ">." << endl;
        }
    } // <-- end of "for" loop
    cout << "End of " << name() << " (ID " << ID() << ").\n";
}


// Class methods below

Mouse* Mouse::returnNewPurebredPtr(int setTo, string name, bool sex) {
    // Previously (in Perl), "setTo" was a two-character string like 'AA' or 'BB', but now it is just an integer.
    
    char sexChr;
    if (Mouse::isMale(sex))   sexChr = 'm';
    else                                sexChr = 'f';
    
    // Note that we ALLOCATE MEMORY here. I don't think we ever free it, though.
    Mouse *purebred = new Mouse(name, kUNDEFINED, kUNDEFINED, NULL, NULL, sexChr); // generation: 0
    // We pass in "-1" for all the IDs that are currently undefined
    
    for (int chr = 0; chr < Chromosome::xChrIndex(); chr++) {
        purebred->getChr(kMALE  , chr)->addPurebredGR(setTo);
        purebred->getChr(kFEMALE, chr)->addPurebredGR(setTo);
    }
    
    if (purebred->isFemale()) {
        // handle two X chromosomes and zero Y chromosomes
        purebred->getChr(kMALE  , Chromosome::xChrIndex())->addPurebredGR(setTo); // X from father
        purebred->getChr(kFEMALE, Chromosome::xChrIndex())->addPurebredGR(setTo); // X from mother
    } else {
        // handle one X chromosome and one Y chromosome
        purebred->getChr(kMALE  , Chromosome::yChrIndex())->addPurebredGR(setTo); // Y from father
        purebred->getChr(kFEMALE, Chromosome::xChrIndex())->addPurebredGR(setTo); // X from mother
     }
     
     return purebred;
}



